'''
The MIT License (MIT)

Copyright © 2023 miltschek.de

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the “Software”), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

from os import listdir, mkdir
from os.path import isfile, basename, exists
from re import A
import zipfile
from xml.dom import minidom
from xml.dom import Node
import xml.parsers.expat
from time import time
import json
import re
import sys

# configure me: location of the .memo files
mypath = '.'

# configure me: filename (and location) of the resulting file
output_filename = 'quillnote_' + str((int(time()))) + '.zip'

def extract_text(node: Node) -> str:
    '''
    Removes any HTML/XML tags and extracts the text content only.
    Caution: depending on your case, you may want to extend this function.
    '''
    if node.nodeType == Node.TEXT_NODE:
        return node.nodeValue
    else:
        result = ''
        for child in node.childNodes:
            child_result = extract_text(child)
            if len(child_result) > 0:
                if len(result) > 0:
                    result = result + '\n'
                result = result + child_result
        return result

def repl(match: re.Match):
    '''
    Converts html escape character codes as they were UTF-16 to a unicode string.
    Expects a match with two groups, each of the denoting a html character escape entity.
    '''
    return (int(match.group(1)).to_bytes(2, byteorder='little') \
        + int(match.group(2)).to_bytes(2, byteorder='little')).decode('utf-16')

def repl_ampersand(match: re.Match):
    '''
    Tries to guess, whether the given ampersand is part of an html-entity
    or an ampersand that should stay as is in the text (but in this case get escaped).
    Expects a match holding an ampersand and a reasonable amount of following characters.
    '''
    if match.group(0).endswith(';'):
        # it's probably an html entity
        # caution: if it fails, need to extend this check to cover all valid entities
        return match.group(0)
    else:
        return '&amp;' + match.group(0)[1:]

quillnote = { 'version': 8, 'notes': [] }

total_files = 0
succeeded_files = 0
total_failures = 0

with zipfile.ZipFile(output_filename, 'w') as output_zip:
    for f in listdir(mypath):
        # analyze files with a .memo extension only
        if isfile(f) and f.endswith('.memo'):
            total_files += 1
            print(f)
            
            try:
                # these files are zip files - open them
                with zipfile.ZipFile(f, 'r') as zf:
                
                    quillnote_entry = { 'id': len(quillnote['notes']) + 1 }
                
                    for entry in zf.infolist():
                        print('-', 'dir' if entry.is_dir() else 'file', entry.filename, entry.compress_size, entry.file_size)
                        
                        # ignore directories
                        if entry.is_dir():
                            continue
                        
                        # process files
                        with zf.open(entry, 'r') as data:
                        
                            # the message itself is stored in an xml file
                            if entry.filename == 'memo_content.xml':
                                doc = None
                                try:
                                    # there is a logical encoding mistake in the memo files
                                    # the message text seems to be encoded in UTF-16
                                    # then, html entities are generated for non-ASCII characters
                                    # (it means, each UTF-16 character gets converted to double &#nnnn; html entity)
                                    # the XML document is UTF-8 encoded
                                    # and the parser complains finding UTF-16 values in an UTF-8 document
                                    # the following two lines correct the issue in a simplified way
                                    # caution: it may happen this trick does not work in your particular case
                                    xml_doc = data.read().decode('utf-8')
                                    xml_doc = re.sub('[&][#]([0-9]+)[;][&][#]([0-9]+)[;]', repl, xml_doc)
                                    doc = minidom.parseString(xml_doc)
                                except Exception as e:
                                    print('- ERROR failed to load message data', f, 'due to', e, file=sys.stderr)
                                    total_failures += 1
                                    continue
                                
                                content_html = None
                            
                                try:
                                    header = doc.firstChild.getElementsByTagName('header')[0]
                                    contents = doc.firstChild.getElementsByTagName('contents')[0]

                                    title = None
                                    favourite = None
                                    uuid = None
                                    createdTime = None

                                    metas = header.getElementsByTagName('meta')
                                    for meta in metas:
                                        for i in range(meta.attributes.length):
                                            attrName = meta.attributes.item(i).name
                                            if attrName == 'title':
                                                title = meta.attributes.item(i).value
                                            elif attrName == 'favourite':
                                                favourite = meta.attributes.item(i).value
                                            elif attrName == 'uuid':
                                                uuid = meta.attributes.item(i).value
                                            elif attrName == 'createdTime':
                                                createdTime = int(meta.attributes.item(i).value)
                                            else:
                                                print('-- unhandled attribute', attrName)
                                    
                                    print('- title', title)
                                    print('- fav', favourite)
                                    print('- uuid', uuid)
                                    print('- createdTime', createdTime)
                                    
                                    # it is possible to have notes without any content
                                    content_html = contents.getElementsByTagName('content')[0].firstChild.nodeValue \
                                        if contents.getElementsByTagName('content')[0].firstChild != None \
                                            else ''

                                except Exception as e:
                                    print('- ERROR failed to extract fields/data from the XML', f, 'due to', type(e), e, file=sys.stderr)
                                    total_failures += 1
                                    continue

                                try:
                                    # the content is an html-formatted document
                                    # then, the content is html-escaped and placed into the XML document
                                    # but any html entities that come from the source document (say, if you used '&' symbol somewhere in the text)
                                    # are escaped only once, generating an error during unescaping
                                    # <p>hi there & here</p> - problem already here!
                                    #   -> &lt;p&gt;hi there &amp; here&lt;/p&gt;
                                    #   -> <xml>...<content>&lt;p&gt;hi there &amp; here&lt;/p&gt;</content>...</xml>
                                    #   and now
                                    #   -> parse XML -> xml/.../content = "<p>hi there & here</p>"
                                    #   -> parse xml/.../content (to get a DOM out of the string) -> error - unexpected '&'
                                    # caution: it may happen you have a message that will get destroyed by this simple fix
                                    # so maybe you need to extend it for your particular case

                                    # the first method does not work properly
                                    # content_html = re.sub('[&][^;]{1,7}[;]?', repl_ampersand, content_html)

                                    # the second method generates excessive escapes, but at least makes the parser not complain
                                    content_html = re.sub('[&]', '&amp;', content_html)

                                    # xml_content seems to contain the same data, but structured in a functional way
                                    # so a to-do-list is a to-do-list instead of text paragraphs
                                    # caution: maybe you need this data in your particular case
                                    # this code ignores it
                                    # xml_content = contents.getElementsByTagName('xml_content')[0].firstChild.nodeValue
                                    
                                    # treat the contents as an XML document
                                    content_dom = minidom.parseString('<xml>' + content_html + '</xml>')
                                    
                                    # extract the text only
                                    content_text = extract_text(content_dom)
                                    
                                    quillnote_entry['title'] = title
                                    quillnote_entry['content'] = content_text
                                    quillnote_entry['creationDate'] = int(createdTime / 1000)
                                    quillnote_entry['modifiedDate'] = quillnote_entry['creationDate']
                                
                                except xml.parsers.expat.ExpatError as e:
                                    print('- ERROR failed to parse the message XML', f, 'due to', e, file=sys.stderr)
                                    print(content_html)
                                    total_failures += 1
                                    continue

                                except Exception as e:
                                    print('- ERROR failed to parse the message XML', f, 'due to', type(e), e, file=sys.stderr)
                                    total_failures += 1
                                    continue
                                
                            else:
                                try:
                                    data = data.read()
                                    # guess the file type from the magic number
                                    filename = str(quillnote_entry['id']) + '_' + basename(entry.filename)
                                    if data.startswith(b'\xff\xd8\xff'):
                                        filename = filename + '.jpg'
                                    elif data.startswith(b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'):
                                        filename = filename + '.png'
                                    else:
                                        print('- WARNING there is a file of unknown type', filename)
                                        filename = filename + '.bin'
                                        
                                    # store files as attachments
                                    output_zip.writestr('media/' + filename, data)
                                    
                                    if not 'attachments' in quillnote_entry:
                                        quillnote_entry['attachments'] = []
                                    
                                    quillnote_entry['attachments'].append({'description': filename, \
                                        'fileName': filename})
                                except Exception as e:
                                    print('- ERROR failed to process attachment', entry.filename, 'due to', e, file=sys.stderr)
                                    total_failures += 1
                                    
                    if 'content' in quillnote_entry \
                        and 'creationDate' in quillnote_entry and quillnote_entry['creationDate'] > 0:
                        quillnote['notes'].append(quillnote_entry)
                        succeeded_files += 1
                    else:
                        print('- ERROR the message', f, 'could not have been added to the set')
                        total_failures += 1

            except Exception as e:
                print('- ERROR failed to unzip the file', f, 'due to', e, file=sys.stderr)
                total_failures += 1

    print('Total files found:', total_files)
    print('Total succeeded files:', succeeded_files)
    print('Total failures:', total_failures)

    output_zip.writestr('backup.json', json.dumps(quillnote))
